"""Web utilities."""

from __future__ import annotations

from contextlib import suppress

__all__: tuple[str, ...] = (
    "ErrorMiddleware",
    "RateLimitMiddleware",
)

import asyncio
from http import HTTPStatus
from typing import TYPE_CHECKING

from aiohttp import ClientResponseError
from loguru import logger
from pydantic import BaseModel, Field

from gongy.errors import (
    AccessDeniedError,
    GongError,
    MalformedRequestError,
    NotFoundError,
    RequestLimitExceededError,
    UnexpectedError,
)

if TYPE_CHECKING:  # pragma: no cover
    from aiohttp import ClientHandlerType, ClientRequest, ClientResponse

    from gongy.errors import GongErrorInfo


class ErrorMiddleware(BaseModel):
    """Middleware to handle errors."""

    errors: set[type[GongError]] = Field(
        default_factory=lambda: {
            MalformedRequestError,
            AccessDeniedError,
            NotFoundError,
            RequestLimitExceededError,
            UnexpectedError,
        }
    )

    async def __call__(
        self,
        req: ClientRequest,
        handler: ClientHandlerType,
    ) -> ClientResponse:
        """Handle errors."""
        try:
            resp = await handler(req)
            resp.raise_for_status()
        except ClientResponseError as e:
            for error_cls in self.errors:
                if error_cls.status == e.status:
                    info: GongErrorInfo = {}
                    with suppress(Exception):
                        info = await resp.json()
                    raise error_cls(
                        message=e.message,
                        request_id=info.get("requestId"),
                        errors=info.get("errors"),
                    ) from e
            raise
        return resp


class RateLimitMiddleware(BaseModel):
    """Middleware to handle rate limiting."""

    retries: int = Field(default=3, gt=0)
    default_delay: float = Field(default=1.0, gt=0)

    async def __call__(
        self,
        req: ClientRequest,
        handler: ClientHandlerType,
    ) -> ClientResponse:
        """Handle rate limiting.

        Raises for status after retries exhausted.
        """
        retries = self.retries
        while True:
            resp = await handler(req)
            if resp.status == HTTPStatus.TOO_MANY_REQUESTS:
                retry_after = float(resp.headers.get("Retry-After", self.default_delay))
                logger.warning(
                    "Rate limit exceeded for request: {method} {url}\n"
                    "Retrying after {retry_after}s ({retries} retries left)",
                    method=req.method,
                    url=req.url,
                    retry_after=retry_after,
                    retries=retries,
                )
                # Handle rate limit exceeded
                if retries <= 0:
                    return resp
                await asyncio.sleep(retry_after)  # Simple backoff strategy
                retries -= 1
                continue
            return resp



# ruff: noqa: ANN401
"""Model utils."""

from __future__ import annotations

__all__: tuple[str, ...] = ("AllTrueModel",)

from types import UnionType
from typing import Any, Self, Union, cast, get_args, get_origin

from pydantic import BaseModel
from pydantic_core import PydanticUndefined


def _is_optional(tp: Any) -> bool:
    origin = get_origin(tp)
    return origin in (UnionType, Union)


def _unwrap_optional(tp: Any) -> Any:
    if not _is_optional(tp):
        return tp
    args = get_args(tp)
    return next((a for a in args if a is not type(None)), None)


def _is_bool_or_optional_bool(tp: Any) -> bool:
    if tp is bool:
        return True
    return _unwrap_optional(tp) is bool if _is_optional(tp) else False


def _is_subclass_of_all_true_model(tp: Any) -> bool:
    try:
        return isinstance(tp, type) and issubclass(tp, AllTrueModel)
    except TypeError:
        return False


def _list_inner_all_true_model(tp: Any) -> Any:
    if get_origin(tp) is list:
        (inner,) = get_args(tp)
        inner = _unwrap_optional(inner) if _is_optional(inner) else inner
        if _is_subclass_of_all_true_model(inner):
            return inner
    return None


def _dict_val_inner_all_true_model(tp: Any) -> Any:
    if get_origin(tp) is dict:
        key_type, val_type = get_args(tp)
        if key_type is not str:
            return None
        val_type = _unwrap_optional(val_type) if _is_optional(val_type) else val_type
        if _is_subclass_of_all_true_model(val_type):
            return val_type
    return None


class AllTrueModel(BaseModel):
    """Base model that defaults all boolean fields to True."""

    @classmethod
    def all(cls: type[Self]) -> Self:
        """Create an instance of the model with all boolean fields set to True."""
        kwargs: dict[str, Any] = {}

        for name, field in cls.model_fields.items():
            ann = field.annotation
            inner_ann = _unwrap_optional(ann) if _is_optional(ann) else ann

            # 1) bool or Optional[bool]
            if _is_bool_or_optional_bool(ann):
                kwargs[name] = True
                continue

            # 2) Nested AllTrueModel (including Optional[Nested])
            if _is_subclass_of_all_true_model(inner_ann):
                inner_ann = cast("type[AllTrueModel]", inner_ann)
                kwargs[name] = inner_ann.all()
                continue

            # 3) list[...] of AllTrueModel (including Optional[list[...] ])
            if list_inner := _list_inner_all_true_model(inner_ann):
                kwargs[name] = [list_inner.all()]
                continue

            # 4) dict[str, ...] of AllTrueModel (including Optional[dict[...] ])
            if dict_val_inner := _dict_val_inner_all_true_model(inner_ann):
                kwargs[name] = {"example": dict_val_inner.all()}
                continue

            # 5) Fallback: default / factory / None if optional
            if field.default is not PydanticUndefined:
                kwargs[name] = field.default
            elif field.default_factory is not None:
                kwargs[name] = field.default_factory()  # type: ignore[call-arg]
            elif _is_optional(ann):
                kwargs[name] = None

        try:
            return cls(**kwargs)  # type: ignore[call-arg]
        except Exception:  # noqa: BLE001
            # If required non-bool/non-nested fields exist without defaults, fall back.
            return cls.model_construct(**kwargs)  # type: ignore[call-arg]



"""Exceptions."""

from __future__ import annotations

__all__: tuple[str, ...] = (
    "AccessDeniedError",
    "CallIDNotFoundError",
    "GongError",
    "MalformedRequestError",
    "NoCallsFoundError",
    "NotFoundError",
    "RequestLimitExceededError",
    "UnexpectedError",
)

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:  # pragma: no cover
    from typing import ClassVar

    from gongy.models import RequestID


class GongErrorInfo(TypedDict, total=False):
    """TypedDict for Gong error information."""

    requestId: RequestID
    errors: list[str]


class GongError(Exception):
    """Base class for all Gong errors."""

    status: ClassVar[int]

    def __init__(
        self,
        message: str | None = None,
        request_id: RequestID | None = None,
        errors: list[str] | None = None,
    ) -> None:
        """Initialize Gong error."""
        super().__init__(message)
        self.request_id = request_id
        self.errors = errors

    def __repr__(self) -> str:
        """Return a string representation of the error."""
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"request_id={self.request_id!r}, errors={self.errors!r})"
        )


class MalformedRequestError(GongError):
    """Exception raised for malformed requests.

    Code: 400
    """

    status = 400


class AccessDeniedError(GongError):
    """Exception raised for access denied errors.

    Code: 401
    """

    status = 401


class NotFoundError(GongError):
    """Exception raised for not found errors.

    Code: 404
    """

    status = 404


class NoCallsFoundError(NotFoundError):
    """Exception raised when no calls are found.

    Code: 404
    """

    status = 404


class CallIDNotFoundError(NotFoundError):
    """Exception raised when a call ID is not found.

    Code: 404
    """

    status = 404


class RequestLimitExceededError(GongError):
    """Exception raised when the request limit is exceeded.

    Code: 429
    """

    status = 429


class UnexpectedError(GongError):
    """Exception raised for unexpected errors.

    Code: 500
    """

    status = 500



"""Custom types."""

from __future__ import annotations

__all__: tuple[str, ...] = (
    "Milliseconds",
    "ms_to_timedelta",
)

from datetime import timedelta
from typing import Annotated

from pydantic import BeforeValidator


def ms_to_timedelta(v: int) -> timedelta:
    """Convert milliseconds to timedelta."""
    return timedelta(milliseconds=v)


Milliseconds = Annotated[
    timedelta,
    BeforeValidator(ms_to_timedelta),
]



"""Gong API."""

from __future__ import annotations

__all__: tuple[str, ...] = ("Gongy",)

from typing import TYPE_CHECKING

from aiohttp import BasicAuth, ClientSession
from pydantic import BaseModel, ConfigDict, Field
from yarl import URL

from gongy.models import (
    CallID,
    CallResponse,
    CallsExpandedResponse,
    CallsRequest,
    CallsResponse,
    CallTranscriptsResponse,
    ContentSelector,
    FilterParams,
    UserID,
)
from gongy.utils.web import (
    ErrorMiddleware,
    RateLimitMiddleware,
)

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import AsyncGenerator
    from datetime import datetime
    from types import TracebackType
    from typing import Self

    from gongy.models import Cursor, WorkspaceID


class Gongy(BaseModel):
    """Gong API client."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    api_key: str
    secret: str

    retries: int = Field(default=3, gt=0)
    delay: float = Field(default=1.0, gt=0)

    base_url: URL = URL("https://api.gong.io")

    raw_session: ClientSession | None = Field(default=None, init=False)

    @property
    def session(self) -> ClientSession:
        """Get the aiohttp session."""
        if self.raw_session is None:
            msg = (
                f"Session not initialized. "
                f"Use 'async with {self.__class__.__name__}(...) as gongy:'"
            )
            raise RuntimeError(msg)
        return self.raw_session

    async def __aenter__(self) -> Self:
        """Enter the async context."""
        self.raw_session = ClientSession(
            auth=BasicAuth(
                login=self.api_key,
                password=self.secret,
            ),
            middlewares=(
                ErrorMiddleware(),
                RateLimitMiddleware(retries=self.retries, default_delay=self.delay),
            ),
        )
        await self.session.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context."""
        await self.session.__aexit__(exc_type, exc_val, exc_tb)
        self.raw_session = None

    @property
    def v2(self) -> URL:
        """Get the v2 API URL."""
        return self.base_url / "v2"

    async def get_calls_page(
        self,
        start: datetime,
        end: datetime,
        workspace: WorkspaceID | None = None,
        cursor: Cursor | None = None,
    ) -> CallsResponse:
        """Get a single page of calls from the Gong API."""
        url = self.v2 / "calls"
        params = {
            "fromDateTime": start.isoformat(),
            "toDateTime": end.isoformat(),
        }
        if workspace is not None:
            params["workspaceId"] = workspace
        if cursor is not None:
            params["cursor"] = cursor
        async with self.session.get(url, params=params) as response:
            data = await response.json()
            return CallsResponse.model_validate(data)

    async def get_calls(
        self,
        start: datetime,
        end: datetime,
        workspace: WorkspaceID | None = None,
    ) -> AsyncGenerator[CallsResponse]:
        """Get calls from the Gong API in batches."""
        cursor: Cursor | None = None
        while True:
            response = await self.get_calls_page(
                start=start,
                end=end,
                workspace=workspace,
                cursor=cursor,
            )
            yield response
            cursor = response.records.cursor
            if cursor is None:
                break

    async def get_call(
        self,
        id: CallID,
    ) -> CallResponse:
        """Get a single call from the Gong API."""
        url = self.v2 / "calls" / id
        async with self.session.get(url) as response:
            data = await response.json()
            return CallResponse.model_validate(data)

    async def get_calls_extensive_page(  # noqa: PLR0913
        self,
        start: datetime | None = None,
        end: datetime | None = None,
        workspace: WorkspaceID | None = None,
        ids: list[CallID] | None = None,
        primary_user_ids: list[UserID] | None = None,
        content_selector: ContentSelector | None = None,
        cursor: Cursor | None = None,
    ) -> CallsExpandedResponse:
        """Get a single page of extensive calls from the Gong API."""
        url = self.v2 / "calls" / "extensive"
        calls_request = CallsRequest(
            cursor=cursor,
            filter=FilterParams(
                from_date_time=start,
                to_date_time=end,
                workspace_id=workspace,
                call_ids=ids,
                primary_user_ids=primary_user_ids,
            ),
            content_selector=content_selector,
        )
        async with self.session.post(
            url,
            json=calls_request.model_dump(
                by_alias=True,
                exclude_none=True,
                mode="json",
            ),
        ) as response:
            data = await response.json()
            return CallsExpandedResponse.model_validate(data)

    async def get_calls_extensive(  # noqa: PLR0913
        self,
        start: datetime | None = None,
        end: datetime | None = None,
        workspace: WorkspaceID | None = None,
        ids: list[CallID] | None = None,
        primary_user_ids: list[UserID] | None = None,
        content_selector: ContentSelector | None = None,
    ) -> AsyncGenerator[CallsExpandedResponse]:
        """Get extensive calls from the Gong API in batches."""
        cursor: Cursor | None = None
        while True:
            response = await self.get_calls_extensive_page(
                start=start,
                end=end,
                workspace=workspace,
                ids=ids,
                primary_user_ids=primary_user_ids,
                content_selector=content_selector,
                cursor=cursor,
            )
            yield response
            cursor = response.records.cursor
            if cursor is None:
                break

    async def get_call_transcripts_page(
        self,
        start: datetime | None = None,
        end: datetime | None = None,
        workspace: WorkspaceID | None = None,
        ids: list[CallID] | None = None,
        cursor: Cursor | None = None,
    ) -> CallTranscriptsResponse:
        """Get a single page of call transcripts from the Gong API."""
        url = self.v2 / "calls" / "transcript"
        calls_request = CallsRequest(
            cursor=cursor,
            filter=FilterParams(
                from_date_time=start,
                to_date_time=end,
                workspace_id=workspace,
                call_ids=ids,
            ),
        )
        async with self.session.post(
            url,
            json=calls_request.model_dump(
                by_alias=True,
                exclude_none=True,
                mode="json",
            ),
        ) as response:
            data = await response.json()
            return CallTranscriptsResponse.model_validate(data)

    async def get_call_transcripts(
        self,
        start: datetime | None = None,
        end: datetime | None = None,
        workspace: WorkspaceID | None = None,
        ids: list[CallID] | None = None,
    ) -> AsyncGenerator[CallTranscriptsResponse]:
        """Get call transcripts from the Gong API in batches."""
        cursor: Cursor | None = None
        while True:
            response = await self.get_call_transcripts_page(
                start=start,
                end=end,
                workspace=workspace,
                ids=ids,
                cursor=cursor,
            )
            yield response
            cursor = response.records.cursor
            if cursor is None:
                break



"""Nox file for automation."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable

import nox

nox.options.default_venv_backend = "uv"

python_versions = (
    (Path(__file__).parent / ".python-versions").read_text().strip().split("\n")
)
python_version = (Path(__file__).parent / ".python-version").read_text().strip()


def install(
    session: nox.Session,
    *,
    groups: Iterable[str],
    root: bool = True,
    extras: bool = False,
) -> None:
    """Install the dependency groups using uv."""
    command = [
        "uv",
        "sync",
        "--frozen",
        "--no-default-groups",
        f"--python={session.virtualenv.location}",
        *(f"--{'group' if root else 'only-group'}={group}" for group in groups),
    ]
    if extras:
        command.append("--all-extras")

    session.run_install(
        *command, env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location}
    )


@nox.session(python=python_version)
def pre_commit(session: nox.Session) -> None:
    """Run pre-commit."""
    install(session, groups=["pre-commit"], root=False)
    session.run(
        "pre-commit",
        "run",
        "--all-files",
        "--show-diff-on-failure",
        "--hook-stage=manual",
    )


@nox.session(python=python_version)
def lock_dependencies(session: nox.Session) -> None:
    """Lock the dependencies."""
    session.run("uv", "lock")


@nox.session(python=python_version)
def lint_files(session: nox.Session) -> None:
    """Lint and fix files."""
    install(session, groups=["linting"], root=False)
    session.run("ruff", "check", ".", "--fix")


@nox.session(python=python_version)
def format_files(session: nox.Session) -> None:
    """Format files."""
    install(session, groups=["linting"], root=False)
    session.run("ruff", "format")


@nox.session(python=python_versions)
def type_check_code(session: nox.Session) -> None:
    """Type-check code."""
    install(session, groups=["typing"], root=True, extras=True)
    # mypy --install-types
    session.run("mypy")


@nox.session(python=python_versions)
def test_code(session: nox.Session) -> None:
    """Test code."""
    install(session, groups=["tests"], root=True, extras=True)
    session.run("pytest")
